/**************************************************************************
 *
 *  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
 *  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
 *  PURPOSE.
 *
 *  Copyright (c) 1993  Microsoft Corporation.  All Rights Reserved.
 * 
 **************************************************************************/
/*
	bkgnd.c
		CreateSpriteWindow -- allocate a SpriteWnd struct
		DeleteSpriteWindow -- frees a SpriteWnd struct, frees sprites
		SetBackground -- assigns a background DIB to a SpriteWnd
			and optionally draws it in the SpriteWnd's hWnd.
*/

#include "global.h"

// CreateSpriteWindow creates a pSprWnd and associates it with an hWnd.
// It does not create any off screen DIBs or DCs; MakeBackground()
// does that. (see global.h for explanation of pSprWnd).
PSPRITEWND PASCAL _export CreateSpriteWindow(HWND hWnd)
{
	PSPRITEWND pSprWnd;

	pSprWnd = (PSPRITEWND) ALLOCATE(sizeof(SPRITEWND));
	if (!pSprWnd)
	{
		Message(0,"Can't allocate SPRITEWND");
		return NULL;
	}
	if (!hWnd)
	{
		Message(0,"CreateSpriteWindow: no hWnd!");
	}
	pSprWnd->hAppInstance = NULL;
	pSprWnd->hwndMain = hWnd;
	pSprWnd->hAccTable = NULL;
	pSprWnd->uiCfSprite = 0;
	pSprWnd->pSpriteList = NULL;
	pSprWnd->pdibBkGnd = NULL;
	pSprWnd->pdibOffScreen = NULL;
	pSprWnd->hdcOffScreen = NULL;
	pSprWnd->pPalClrTable = NULL;
	pSprWnd->uiTimer = 0;
	pSprWnd->hpalCurrent = NULL;
	pSprWnd->bShowUpdateRects = NULL;
	pSprWnd->bAutoUpdate = NULL;
	return pSprWnd;
}

// DeleteSP() frees a pSprWnd and its associated off-screen
// DC and DIB.  (There is no UnMakeBackground, see?).
void PASCAL _export DeleteSpriteWindow(PSPRITEWND pSprWnd)
{
	if (!pSprWnd)
	{
		Message(0,"DeleteSpriteWindow: no pSprWnd!!");
		return;
	}
	if (pSprWnd->hdcOffScreen)
	{
		DeleteDC(pSprWnd->hdcOffScreen);
		pSprWnd->hdcOffScreen = NULL;
	}
	if (pSprWnd->pdibOffScreen)
	{
		DeleteDIB(pSprWnd->pdibOffScreen);
		pSprWnd->pdibOffScreen = NULL;
	}
	DeleteSpriteList(pSprWnd);
	if (pSprWnd->hpalCurrent) DeleteObject(pSprWnd->hpalCurrent);
}

// SetBackground takes the pdibBkGnd that you pass it and
// makes it the background in the hWnd you specified when you
// created the pSprWnd.  It doesn't actually show up unless you
// set bRedraw to true.
//
// This routine creates an off-screen DIB of the same size as the
// background DIB (by calling CreateCompatibleDIB in DIB.C) and a
// DC using the DIB driver that uses the off-screen DIB.  It also
// sets up a pallette based on the DIB you provide (by calling
// CreateDIBPalette in palette.c).  All the sprites should use
// this pallette.
void PASCAL _export SetBackground(PSPRITEWND pSprWnd, PDIB pdibBkGnd, BOOL bRedraw)
{
	LPWORD pIndex;
	int i;
	
	if (!pSprWnd)
	{
		Message(0,"SetBackground: pSprWnd is NULL");
		return;
	}
	if (!pdibBkGnd)
	{
		Message(0,"SetBackground: pdibBkGnd is NULL");
		return;
	}
	// set background DIB
	dprintf2("SetBackground(%ld)", pdibBkGnd);
	pSprWnd->pdibBkGnd = pdibBkGnd;
	if (!pdibBkGnd)
	{
		return;
	}
	// set pallette
	if (pSprWnd->hpalCurrent) DeleteObject(pSprWnd->hpalCurrent);
	pSprWnd->hpalCurrent = CreateDIBPalette(pSprWnd->pdibBkGnd);
	SetSysPalColors(pSprWnd->hpalCurrent);
	// if off-screen DC exists, delete it
	if (pSprWnd->hdcOffScreen)
	{
		DeleteDC(pSprWnd->hdcOffScreen);
		pSprWnd->hdcOffScreen = NULL;
	}
	DeleteDIB(pSprWnd->pdibOffScreen);
	// create off-screen DIB same size as this one
	pSprWnd->pdibOffScreen = CreateCompatibleDIB(pSprWnd->pdibBkGnd);
	if (!pSprWnd->pdibOffScreen)
	{
		dprintf1("Failed to create off-screen DIB");
		return;
	}
	// create off-screen DC using DIB driver
	pSprWnd->hdcOffScreen = CreateDC("DIB", NULL, NULL, (LPSTR)pSprWnd->pdibOffScreen);
	if (!pSprWnd->hdcOffScreen)
	{
		dprintf1("Failed to create off-screen DC");
	}
	else
	{
		dprintf3("Off-screen DC created");
	}
	// if debugging, spit out info
	#ifdef DEBUG
	{
		WORD wCaps;
		wCaps = GetDeviceCaps(pSprWnd->hdcOffScreen, RASTERCAPS);
		if (wCaps & RC_DI_BITMAP)
		{
			dprintf3("DIB driver DC supports GetDIBits and SetDIBits");
		}
		else
		{
			dprintf3("DIB driver DC doesn't support GetDIBits or SetDIBits");
		}
		if (wCaps & RC_DIBTODEV)
		{
			dprintf3("DIB driver DC supports SetDIBitsToDevice");
		}
		else
		{
			dprintf3("DIB driver DC doesn't support SetDIBitsToDevice");
		}
		if (wCaps & RC_STRETCHDIB)
		{
			dprintf3("DIB driver DC supports StretchDIBits");
		}
		else
		{
			dprintf3("DIB driver DC doesn't support StretchDIBits");
		}
		wCaps = GetDeviceCaps(pSprWnd->hdcOffScreen, CAPS1);
		if (wCaps & C1_TRANSPARENT)
		{
			dprintf3("DIB driver DC supports transparency");
		}
		else
		{
			dprintf3("DIB driver DC doesn't support transparency");
		}
	}
	#endif
	// Create a new 1:1 lookup table so we can use DIB_PAL_COLORS
	// when we call StretchDIBits to blt data from the off-screen
	// DC to the window DC.
	if (!pSprWnd->pPalClrTable)
	{
		pSprWnd->pPalClrTable = (LPBITMAPINFO) ALLOCATE(sizeof(BITMAPINFOHEADER)
			+ 256 * sizeof(WORD));
		if (!pSprWnd->pPalClrTable)
		{
			dprintf1("Out of memory for 1:1 clr table");
			return;
		}
	}
	// Set up the table to match the off-screen DIB
	// by copying the BITMAPINFOHEADER and building a 1:1 color table
	_fmemcpy(pSprWnd->pPalClrTable,
		pSprWnd->pdibOffScreen,
		sizeof(BITMAPINFOHEADER));
	pIndex = (LPWORD)((LPSTR)(pSprWnd->pPalClrTable) + sizeof(BITMAPINFOHEADER));
	for (i=0; i<256; i++)
	{
		*pIndex++ = (WORD) i;
	}
	// Draw the background to the off-screen DIB
	// and update the screen if required
	Redraw(pSprWnd, NULL, bRedraw);
}
