/**************************************************************************
 *
 *  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
 *  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
 *  PURPOSE.
 *
 *  Copyright (c) 1993  Microsoft Corporation.  All Rights Reserved.
 * 
 **************************************************************************/
/*
	sprite.c
		CreateSprite -- allocates a sprite and initializes it
		DeleteSprite -- undoes CreateSprite
		DeleteSpriteList -- deletes all the sprites in a SpriteWindow
		SpriteHitTest -- see if x and y are on a sprite, if so which
		GetSpriteRect -- creates a Windows RECT for sprite
		SetSpritePosition -- move a sprite
		SetSpriteZOrder -- move a sprite in front of or behind others
		SetSpriteDIB -- change the appearance of a sprite
		SetSpriteVisible -- make sprite visible or invisible
*/

#include "global.h"

// CreateSprite allocates a sprite and adds
// it to the sprite list for the given SpriteWindow.
// The sprite is positioned at the given x,y,z (z is
// for determining which sprites appear on top of which
// others) and is optionally drawn on the screen
PSPRITE PASCAL _export CreateSprite(PSPRITEWND pSprWnd, PDIB pSprDib, int initx, int inity, int initz, BOOL bRedraw)
{
	PSPRITE pSprite;
	RECT rc;

	if (!pSprWnd)
	{
		Message(0,"CreateSprite: pSprWnd is NULL");
		return NULL;
	}
	dprintf2("LoadSprite()");

	// Allocate memory for the SPRITE info
	pSprite = (PSPRITE) ALLOCATE(sizeof(SPRITE));
	if (!pSprite)
	{
		dprintf1("No memory for sprite");
		return NULL;
	}
	// assign DIB image
	pSprite->pDIB = pSprDib;
	if (!pSprite->pDIB)
	{
		Message(0,"CreateSprite: pDIB is NULL");
		if (pSprite) FREE(pSprite);
		return NULL;
	}
	// initialize
	pSprite->width = (int) DIB_WIDTH(pSprite->pDIB);
	pSprite->height = (int) DIB_HEIGHT(pSprite->pDIB);
	pSprite->x = initx;
	pSprite->y = inity;
	pSprite->z = initz;
	pSprite->pNext = NULL;
	pSprite->pPrev = NULL;
	pSprite->bVisible = TRUE;
	dprintf4("Sprite is %u by %u", pSprite->width, pSprite->height);

	// Map the color table so that it matches the one in the
	// background DIB
	ASSERT(pSprWnd->pdibBkGnd);
	MapDIBColorTable(pSprite->pDIB, pSprWnd->pdibBkGnd);

	// get the index and color of the top left pixel
	pSprite->bTopLeft = GetDIBPixelValue(pSprite->pDIB, 0, 0);
	pSprite->rgbTopLeft = GetDIBPixelColor(pSprite->pDIB, 0, 0);
	dprintf4("Sprite top left: %2.2XH, %8.8lXH",
		pSprite->bTopLeft,
		pSprite->rgbTopLeft);

	// Add it to the sprite list
	if (pSprWnd->pSpriteList)
	{
		pSprWnd->pSpriteList->pPrev = pSprite;
		pSprite->pNext = pSprWnd->pSpriteList;
	}
	pSprWnd->pSpriteList = pSprite;

	// Set the zorder
	SetSpriteZOrder(pSprWnd, pSprite, initz, NO_UPDATE);
	GetSpriteRect(pSprite, &rc);

	// Redraw the sprite set
	Redraw(pSprWnd, &rc, bRedraw);
	return pSprite;
}

// Delete sprite frees memory associated with a sprite,
// and remove it from its sprite list. We DO NOT delete
// the sprite's DIB, so this DIB can be used in other sprites.
void PASCAL _export DeleteSprite(PSPRITEWND pSprWnd, PSPRITE pSprite)
{
	PSPRITE pPrev, pNext;

	if (!pSprWnd)
	{
		Message(0,"DeleteSprite: pSprWnd is NULL");
		return;
	}
	if (!pSprWnd)
	{
		Message(0,"DeleteSprite: pSprite is NULL");
		return;
	}
	pPrev = pSprite->pPrev;
	pNext = pSprite->pNext;

	if (pPrev)
	{
		pPrev->pNext = pNext;
	}
	else
	{
		pSprWnd->pSpriteList = pNext;
	}
	if (pNext)
	{
		pNext->pPrev = pPrev;
	}
}

// Delete all sprites in a SpriteWindow.
// Note that NO DIBS ARE DELETED by this.
void DeleteSpriteList(PSPRITEWND pSprWnd)
{
	PSPRITE pSprite, pNext;

	pSprite = pSprWnd->pSpriteList;
	while (pSprite)
	{
		pNext = pSprite->pNext;
		DeleteSprite(pSprWnd, pSprite);
		pSprite = pNext;
	}
	pSprWnd->pSpriteList = NULL;
}

// Check to see if the given x,y coordinates are
// on top of a sprite.  First we check to see if
// the sprite rect is intersected, and if it is,
// we look at the sprite's bits to see if we hit
// a transparent or non-transparent pixel.
PSPRITE PASCAL _export SpriteHitTest(PSPRITEWND pSprWnd, int x, int y)
{
	PSPRITE pSprite;
	int dx, dy;

	if (!pSprWnd)
	{
		Message(0,"SpriteHitTest: pSprite is NULL");
		return NULL;
	}
	pSprite = pSprWnd->pSpriteList;
	while (pSprite)
	{
		// Test if the click is inside the sprite rectangle
		if ((x > pSprite->x)
			&& (x < pSprite->x + (int) pSprite->width)
			&& (y > pSprite->y)
			&& (y < pSprite->y + (int) pSprite->height))
		{
			dprintf4("Hit is in sprite rect");

			// See if this point is transparent by testing to
			// see if the pixel value is the same as the top
			// left corner value.  Note that top left of the
			// image is bottom left in the DIB.
			dx = x - pSprite->x;
			dy = y - pSprite->y;
			if (GetDIBPixelValue(pSprite->pDIB, dx, dy)
				!= pSprite->bTopLeft)
			{
				break;
			}
		}
		pSprite = pSprite->pNext;
	}
	return pSprite;
}

// GetSpriteRect converts a sprites x, y, width, and height
// values into a Windows RECT structure for use with
// Windows APIs
void GetSpriteRect(PSPRITE pSprite, LPRECT prcSprite)
{
	prcSprite->left = pSprite->x;
	prcSprite->right = pSprite->x + pSprite->width;
	prcSprite->top = pSprite->y;
	prcSprite->bottom = pSprite->y + pSprite->height;
}

// SetSpritePosition moves a sprite to a given x,y coordinate.
// Both the old position and the new are invalidated, and are
// optionally redrawn.  It is assumed that the new position
// will intersect the old.
void PASCAL _export SetSpritePosition(PSPRITEWND pSprWnd, PSPRITE pSprite, int x, int y, BOOL bUpdate)
{
	RECT rcOld, rcNew, rcChange;

	dprintf4("SetSpritePosition(%u,%u)", x, y);
	if (!pSprWnd)
	{
		Message(0,"SetSpritePosition: pSprWnd is NULL");
		return;
	}
	if (!pSprWnd)
	{
		Message(0,"SetSpritePosition: pSprite is NULL");
		return;
	}

	// Get the current sprite position
	GetSpriteRect(pSprite, &rcOld);

	// Update the position and get the new rectangle
	pSprite->x = x;
	pSprite->y = y;
	GetSpriteRect(pSprite, &rcNew);

	// Update the change rectangle to include the old and the new
	// positions of the sprite
	CopyRect(&rcChange, &rcOld);
	UnionRect(&rcChange, &rcChange, &rcNew);
	if (bUpdate != NO_UPDATE)
	{
		Redraw(pSprWnd, &rcChange, bUpdate);
	}
}

// SetSpriteZOrder moves a sprite above or below other sprites.
// This requires changing its position in the linked list of sprites.
// The rect occupied by the sprite is optionally repainted.
void PASCAL _export SetSpriteZOrder(PSPRITEWND pSprWnd, PSPRITE pSprite, WORD z, BOOL bUpdate)
{
	RECT rcChange;
	PSPRITE pNew, pLast;

	dprintf4("SetSpriteZOrder(%u)", z);
	if (!pSprWnd)
	{
		Message(0,"SetSpriteZOrder: pSprWnd is NULL");
		return;
	}
	if (!pSprWnd)
	{
		Message(0,"SetSpriteZOrder: pSprite is NULL");
		return;
	}
	ASSERT(pSprWnd->pSpriteList);
	pSprite->z = z;

	// Get the current sprite position
	GetSpriteRect(pSprite, &rcChange);

	// Unlink the sprite from the list
	if (pSprite->pPrev)
	{
		pSprite->pPrev->pNext = pSprite->pNext;
	}
	else
	{
		pSprWnd->pSpriteList = pSprite->pNext;
	}
	if (pSprite->pNext)
	{
		pSprite->pNext->pPrev = pSprite->pPrev;
	}
	pSprite->pNext = pSprite->pPrev = NULL;

	// Walk down the list (from the top (front) until we find
	// a sprite with higher z order or the end of the list
	pNew = pLast = pSprWnd->pSpriteList;
	while (pNew)
	{
		if (pNew->z > pSprite->z) break;
		pLast = pNew;
		pNew = pNew->pNext;
	}
	if (!pNew)
	{
		// There is no sprite further back than this so put
		// this one on the end
		if (pLast)
		{
			pLast->pNext = pSprite;
		}
		else
		{
			pSprWnd->pSpriteList = pSprite;
		}
		pSprite->pPrev = pLast;
	}
	else
	{
		// pNew points the the one we want after this one
		// so insert the new sprite before it
		if (pNew->pPrev)
		{
			pSprite->pNext = pNew;
			pSprite->pPrev = pNew->pPrev;
			pNew->pPrev->pNext = pSprite;
			pNew->pPrev = pSprite;
		}
		else
		{
			pSprite->pNext = pNew;
			pSprWnd->pSpriteList = pSprite;
			pNew->pPrev = pSprite;
		}
	}
	// Redraw the changes if required
	Redraw(pSprWnd, &rcChange, bUpdate);
}

// SetSpriteDIB changes the DIB associated with a sprite to a new
// one, and optionally draws it on the screen.  Note that the old
// DIB is NOT deleted.  This enables one to cycle through a number
// of DIBs to animate a sprite.
//
// This routine might have a bug because it doesn't call MapDIBColorTable
// (calling MapDIBColorTable seemed to produce a bug so I removed it)
void PASCAL _export SetSpriteDIB(PSPRITEWND pSprWnd, PSPRITE pSprite, PDIB pSprDib, BOOL bRedraw)
{
	RECT rcOld, rcNew, rcChange;

	dprintf2("ChangeSpriteImage()");
	if (!pSprWnd)
	{
		Message(0,"SetSpriteDIB: pSprWnd is NULL");
		return;
	}
	if (!pSprWnd)
	{
		Message(0,"SetSpriteDIB: pSprite is NULL");
		return;
	}

	// get old rect
	GetSpriteRect(pSprite, &rcOld);

	// assign new image
	pSprite->pDIB = pSprDib;
	if (!pSprite->pDIB)
	{
		if (pSprite) FREE(pSprite);
		Message(0,"SetSpriteDIB: pDIB is NULL");
		return;
	}

	// get new size
	pSprite->width = (int) DIB_WIDTH(pSprite->pDIB);
	pSprite->height = (int) DIB_HEIGHT(pSprite->pDIB);

	// get new rect
	GetSpriteRect(pSprite, &rcNew);
	CopyRect(&rcChange, &rcOld);

	// combine old and new rects
	UnionRect(&rcChange, &rcChange, &rcNew);
	dprintf4("Sprite is %u by %u", pSprite->width, pSprite->height);
	ASSERT(pSprWnd->pdibBkGnd);

	// assign new transparency color
	pSprite->bTopLeft = GetDIBPixelValue(pSprite->pDIB, 0, 0);
	pSprite->rgbTopLeft = GetDIBPixelColor(pSprite->pDIB, 0, 0);
	dprintf4("Sprite top left: %2.2XH, %8.8lXH",
		pSprite->bTopLeft,
		pSprite->rgbTopLeft);

	// redraw the sprite if requested
	Redraw(pSprWnd, &rcChange, bRedraw);
}

// This was added to provide a quick way to make a sprite
// invisible, rather than setting it to an empty image or
// moving it off the display.
void PASCAL _export SetSpriteVisible(PSPRITEWND pSprWnd, PSPRITE pSprite, BOOL bVisible, BOOL bRedraw)
{
	RECT rcChange;

	dprintf2("ChangeSpriteVisible()");
	if (!pSprWnd)
	{
		Message(0,"SetSpriteVisible: pSprWnd is NULL");
		return;
	}
	if (!pSprWnd)
	{
		Message(0,"SetSpriteVisible: pSprite is NULL");
		return;
	}

	// get rect to redraw
	GetSpriteRect(pSprite, &rcChange);

	// set visible property to given state
	pSprite->bVisible = bVisible;
	dprintf4("Sprite is %u by %u", pSprite->width, pSprite->height);
	ASSERT(pSprWnd->pdibBkGnd);
	
	// redraw if requested
	Redraw(pSprWnd, &rcChange, bRedraw);
}


